#!/usr/bin/env python3
"""
This example demonstrates how one can work with boolean values
and additonal kwargs.


./examples/uftpd.py  server --help
usage: uftpd.py server [-h] [--version] [--syslog] [--foreground]
                       [--level LEVEL] [--opts OPTS]

optional arguments:
  -h, --help     show this help message and exit
  --version      Show the program version and exit (default: False)
  --syslog       Use syslog, even if running in foreground (default: False)
  --foreground   Run in foreground, do not detach from controlling terminal
                 (default: False)
  --level LEVEL  set the verbosity level: none, err, info, notice, debug
                 (default: 2)
  --opts OPTS    Additional options loaded from JSON (default: None)

$ ./examples/uftpd.py --opts='{"ftp": 21}'
serving FTP on port 21

$ ./examples/uftpd.py --opts='{"ftp": 21, "foo": "bar"}'
Unknown option foo given
"""

import sys

from mach import mach1


@mach1
class uFTPD:

    default = 'server'

    def server(self, level: int=2, foreground: bool=False,
               syslog: bool=False, version: bool=False, **opts):
        """No nonsense TFTP/FTP Server

        level - set the verbosity level: none, err, info, notice, debug
        foreground - Run in foreground, do not detach from controlling terminal
        syslog - Use syslog, even if running in foreground
        version - Show the program version and exit
        opts - ftp=PORT, tftp=PORT


        """
        if version:
            print("uftpd version 0.1")
            sys.exit(0)

        for key in opts:
            if key not in ['ftp', 'tftp']:
                print("Unknown option %s given" % key)
                sys.exit(1)

        print("serving FTP on port %d" % opts['ftp'])


if __name__ == '__main__':
    uFTPD().run()



#!/usr/bin/env python3
# demonstrate how to enable only shell

import sys

from mach import mach2

@mach2
class FTPClient:

    def login(self, user: str, password: str):
        """login to the FTP server"""

        self.stdout.write("Login success ...\n")

    def connect(self, host: str, port: int=21, **opts):
        """connect to FTP host

        host - the host IP or fqdn
        port - the port listening to FTP
        """
        self.stdout.write("Connected to %s:%s\n" % (host, port))

        if opts:
            self.login(**opts)

    def ls(self, dir: str="/"):
        """list files in dir"""
        self.stdout.write("Files in %s\n" % dir)

    def exit(self):
        """exit the program"""
        sys.exit(0)

if __name__ == '__main__':
    ftpclient = FTPClient()
    ftpclient.intro = 'Welcome to the lftp client. Type help or ? to list commands.\n'
    ftpclient.prompt = 'lftp > '

    ftpclient.cmdloop()



#!/usr/bin/env python3
"""
This example demonstrate how decorating with `mach1`
turns your class into a command line application.


   $ ./examples/calc.py add 2 4
   4 + 2 => 6
   $ ./examples/calc.py div 2 4
   4 / 2 => 2
"""

from mach import mach1


@mach1
class Calculator:

    def add(self, a: int, b: int):
        """adds two numbers and prints the result"""
        print("%s + %s => %d" % (a, b, int(a) + int(b)))

    def div(self, a: int, b: int):
        """divide one number by the other"""
        print("%s / %s => %d" % (a, b, int(a) // int(b)))



if __name__ == '__main__':

    calc = Calculator()
    calc.run()



#!/usr/bin/env python3
"""
This example demonstrates how one can define a default subcommand
which is invoked if no other subcommand is given::

   $ ./examples/greet.py -h
   usage: greet.py [-h] {greet,part} ...

   positional arguments:
     {greet,part}  commands
       greet       Greets a user one or more times
       part        Politely part from a user

   optional arguments:
  -h, --help    show this help message and exit
   $ ./examples/greet.py greet --name Tom
   Hello Tom
   $ ./examples/greet.py
   Your name: Tom
   Hello Tom
   $

This example also shows how you can use a special format for the method
docstring to expand the help of subcommands options::

   $ ./examples/greet.py greet -h
usage: greet.py greet [-h] [--name NAME] [--count COUNT]

optional arguments:
  -h, --help     show this help message and exit
  --name NAME    the name of the user to greet (default: )
  --count COUNT  the number of times to greet the user (default: 1)
"""

from mach import mach1


@mach1
class Hello:

    default = 'greet'

    def greet(self, count: int=1, name: str=""):
        """Greets a user one or more times

        count - the number of times to greet the user
        name - the name of the user to greet
        """

        if not name:
            name = input('Your name: ')

        for c in range(count):
            print("Hello %s" % name)

    def part(self):
        """Politely part from a user"""
        print("It was nice to meet you!")


if __name__ == '__main__':
    Hello().run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `mach` module."""
import argparse
import cmd
import os
import sys

from io import StringIO
from unittest import mock


import pytest


from mach import mach1, mach2


import examples
from examples.greet import Hello
from examples.calc import Calculator
from examples.calc2 import Calculator as Calc2
from examples.uftpd import uFTPD
from examples.lftp import FTPClient


def test_hello():

    hello = Hello()

    assert isinstance(hello.parser, argparse.ArgumentParser)


def test_calc():

    calc = Calculator()

    assert isinstance(calc.parser, argparse.ArgumentParser)


def test_calc2():

    calc = Calc2()

    assert isinstance(calc.parser, argparse.ArgumentParser)
    assert isinstance(calc, cmd.Cmd)

    # https://stackoverflow.com/q/34500249/492620
    with mock.patch('sys.stdout', new=StringIO()) as fakeOutput:
        calc.add(1, 3)
        assert '1 + 3 => 4' == fakeOutput.getvalue().strip()

    with mock.patch('sys.stdout', new=StringIO()) as fakeOutput:
        calc.do_add(1, 3)
        assert '1 + 3 => 4' == fakeOutput.getvalue().strip()

    with mock.patch('sys.stdout', new=StringIO()) as fakeOutput:
        calc.onecmd('add 1 2')
        assert '1 + 2 => 3' == fakeOutput.getvalue().strip()


def test_calc2_no_interactive():
    calc = Calc2()

    # this is ./examples/calc2.py add 2 4
    with mock.patch('sys.stdout', new=StringIO()) as fakeOutput:
        calc._run1(["add", "2", "4"])
        assert '4 + 2 => 6' == fakeOutput.getvalue().strip()



def test_uftpd():

    uftpd = uFTPD()
    try:
        uftpd.run(['server', '--version'])
    except SystemExit:
        pass

    opts="""--opts={"ftp": 21, "foo": "bar"}"""

    try:
        uftpd.run(['server', opts])
    except SystemExit:
        pass


@pytest.mark.parametrize("input,output", [
     ("connect foo.example.com 21", "Connected to foo.example.com:21"),
     ("connect foo.example.com", "Connected to foo.example.com:21"),
     ("connect foo.bar.com port=2121", "Connected to foo.bar.com:2121"),
     ('connect ftp.example.com 21 opts={"user": oz123, "password":s3kr35}', "Could not parse JSON in opts"),
     ("""connect ftp.example.com 21 opts=\'{"user": "oz123", "password":"s3kr35"}\'""",
"""Connected to ftp.example.com:21
Login success ..."""),
     ("login foo s3kr35", "Login success ..."),
     ('""', '*** Unknown syntax: ""'),
     ("login foo bar bla", "*** Unknown syntax: login foo bar bla"),
     ("moo", "*** Unknown syntax: moo"),
     ("ls", "Files in /"),
     ("ls /foo/", "Files in /foo/"),
     ("help", """Documented commands (type help <topic>):
========================================
connect  exit  help  login  ls"""),
     ("connect foo=21", "Unknown option foo"),
     ("""connect foo.example.com 21 opts='{"user": "oz123", "password": "s3kr35"}""",
      """connect foo.example.com 21 opts='{"user": "oz123", "password": "s3kr35"}: No closing quotation"""),
])
def test_lftp(input, output):
    ftpc = FTPClient(stdout=StringIO())

    ftpc.onecmd(input)
    assert ftpc.stdout.getvalue().strip() == output




"""
This file is part of m.a.c.h.

m.a.c.h is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

m.a.c.h is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with m.a.c.h.  If not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import inspect
import json
import pkg_resources
import shlex

from cmd import Cmd
from itertools import filterfalse, tee

try:
    __version__ = pkg_resources.get_distribution('pwman3').version
except pkg_resources.DistributionNotFound:  # pragma: no cover
    __version__ = "0.0.1"


def partition(pred, iterable):
    'Use a predicate to partition entries into false entries and true entries'
    # partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
    t1, t2 = tee(iterable)
    return list(filterfalse(pred, t1)), list(filter(pred, t2))


class DefaultSubcommandArgParse(argparse.ArgumentParser):
    # https://stackoverflow.com/a/37593636/492620
    __default_subparser = None

    def set_default_subparser(self, name):
        self.__default_subparser = name

    def _parse_known_args(self, arg_strings, *args, **kwargs):
        in_args = set(arg_strings)
        d_sp = self.__default_subparser
        if d_sp is not None and not {'-h', '--help'}.intersection(in_args):
            for x in self._subparsers._actions:
                subparser_found = (
                    isinstance(x, argparse._SubParsersAction) and
                    in_args.intersection(x._name_parser_map.keys())
                )
                if subparser_found:
                    break
            else:
                # insert default in first position, this implies no
                # global options without a sub_parsers specified
                arg_strings = [d_sp] + arg_strings
        return super(DefaultSubcommandArgParse, self)._parse_known_args(
            arg_strings, *args, **kwargs
        )


class Mach(Cmd):

    def onecmd(self, line):
        cmd, arg, line = self.parseline(line)
        if not line:
            return self.emptyline()
        if cmd is None:
            return self.default(line)
        self.lastcmd = line
        if line == 'EOF':
            self.lastcmd = ''
        if cmd == '':
            return self.default(line)

        try:
            arg = shlex.split(arg)
        except ValueError as e:
            self.stdout.write(line + ": %s\n" % e)
            return

        arg, args_with_val = partition(lambda x: "=" in x, arg)
        di = {}

        try:
            func = getattr(self, 'do_' + cmd)
        except AttributeError:
            # when a method is not found
            return self.default(line)

        sig = inspect.getfullargspec(func)
        for item in args_with_val:
            name, val = item.split('=')
            if name != sig.varkw and name not in sig.args[1:]:
                self.stdout.write("Unknown option %s\n" % name)
                return
            di[name] = val


        if sig.varkw and sig.varkw in di:
            try:
                kwargs = json.loads(di.pop(sig.varkw))
                di.update(kwargs)
            except json.decoder.JSONDecodeError:
                self.stdout.write("Could not parse JSON in %s\n" % sig.varkw)
                return

        try:
            return func(*arg, **di)
        except ValueError:
            # when a method is wrongly used
            return self.default(line)

        except TypeError as e:
            if e.args[0].endswith(
                "missing 1 required positional argument: 'arg'"):
                return func("")
            else:
                return self.default(line)


        # Cmd methods except one arg
        #try:
        #    return func("")
        #except TypeError:
        #    return self.default(line)


_supported_types = {'str': str, 'float': float, 'int': int}


def parse_docs(docstring):
    """
    Parse documentation string and create a help string
    """

    doc = docstring.split("\n")
    doc_dict = {'cmd': doc[0]}
    if len(doc) > 1:
        doc = {k: v for k, v in
               (item.split(' - ') for item in
                list(filter(lambda x: x, doc))[1:])}
        doc_dict.update(doc)
    return doc_dict


def add_parsers(name, function, doc, sig, subparsers):

    subp = subparsers.add_parser(
        name, formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help=doc['cmd'])

    idx_args_with_defaults = len(sig.defaults) if sig.defaults else 0

    if sig.defaults:
        _defaults = list(reversed(sig.defaults))

    for idx, val in enumerate(reversed(sig.args[1:])):
        subpargs, opts = [], {}
        opts['help'] = doc.get(val, '')
        type_ = sig.annotations.get(val)
        if type_ and type_.__name__ == 'bool':
            opts['action'] = "store_true"
        elif type_ and type_.__name__ in _supported_types:
                opts['type'] = _supported_types[type_.__name__]
        if idx_args_with_defaults and idx < idx_args_with_defaults:
            opts['default'] = _defaults[idx]
            subpargs.append("--" + val)
        else:
            subpargs.append(val)

        subp.add_argument(*subpargs, **opts)
    if sig.varkw:
        subp.add_argument("--" + sig.varkw,
                          help="Additional options loaded from JSON")

    return name, function, doc

def create_helper(doc, name):
    return lambda name: print(doc)


def _mach(kls, add_do=False):

    if hasattr(kls, 'default'):
        parser = DefaultSubcommandArgParse(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    else:
        parser = argparse.ArgumentParser(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    subparsers = parser.add_subparsers(help='commands', dest="cmd")

    if add_do:
        do_kls = type(kls.__name__, (Mach, kls), {})
    for (name, function) in inspect.getmembers(kls,
                                               predicate=inspect.isfunction):
        _d = inspect.getdoc(function)
        doc = parse_docs(_d)
        sig = inspect.getfullargspec(function)
        add_parsers(name, function, doc, sig, subparsers)

        if add_do:
            setattr(do_kls, "do_%s" % name, function)
            setattr(do_kls, "help_%s" % name, create_helper(_d, name))


    if hasattr(kls, 'default'):
        parser.set_default_subparser(kls.default)

    if add_do:
        kls = do_kls

    kls.parser = parser
    return kls


def _run1(inst, args=None):
    p = inst.parser.parse_args(args=args)

    if p.cmd:
        func_args_kwargs = inspect.getfullargspec(getattr(inst, p.cmd))
        args = func_args_kwargs.args
        args.pop(args.index("self"))
        if func_args_kwargs.varkw:
            kwargs = getattr(p, func_args_kwargs.varkw)
            if kwargs:
                kwargs = json.loads(kwargs)
            else:
                kwargs = {}
            func = getattr(inst, p.cmd)
            func(*(getattr(p, arg) for arg in args), **kwargs)
        else:
            func = getattr(inst, p.cmd)
            func(*(getattr(p, arg) for arg in args))
            return True


def _run2(inst):  # pragma: no coverage
    if not inst._run1():
        inst.cmdloop()


def mach1(kls):  # pragma: no coverage
    kls = _mach(kls)
    kls.run = _run1
    return kls


def mach2(kls): #  pragma: no coverage
    kls = _mach(kls, add_do=True)
    kls._run1 = _run1
    kls.run = _run2
    return kls



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

# with open('requirements.txt') as r:
#     requirements = r.readlines()
#     requirements = [r.split(' ', 1) for r in requirements][0]
requirements = []

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    setup_requires=['pbr'],
    pbr=True,
)



#!/usr/bin/env python3
"""
This example demonstrate how decorating with `mach1`
turns your class into a command line application, which also
has it's own REPL::

   $ ./examples/calc2.py add 4 1
   1 + 4 => 5

   $ ./examples/calc2.py
   Welcome to the calc shell. Type help or ? to list commands.

   calc2 > add 4 1
   4 + 1 => 5
   calc2 > exit
   Come back soon ...
"""

import sys

from mach import mach2


@mach2
class Calculator:

    def add(self, a: int, b: int):
        """adds two numbers and prints the result"""
        print("%s + %s => %d" % (a, b, int(a) + int(b)))

    def div(self, a: int, b: int):
        """divide one number  by the other"""
        print("%s / %s => %d" % (a, b, int(a) // int(b)))

    def exit(self):
        """exist to finish this session"""
        print("Come back soon ...")
        sys.exit(0)


if __name__ == '__main__':
    calc = Calculator()
    calc.intro = 'Welcome to the calc shell. Type help or ? to list commands.\n'
    calc.prompt = 'calc2 > '
    calc.run()
